unit uSquarifiedTreeMap;

interface

uses
  System.Classes,
  System.Types,
  System.Generics.Collections,
  System.SysUtils,
  System.IOUtils,
  Vcl.Graphics,
  Vcl.Dialogs;

type
  TFileNode = class
  private
    FFileName: string;
    FSize: UInt64;
    FDepth: Integer;
  public
    Bloc: TRect;
    Children: TObjectList<TFileNode>;
    constructor Create(const AFileName: string; ASize: Int64; ADepth: Integer);
    destructor Destroy; override;

    procedure ShowArbo;

    property Filename: string read FFilename;
    property Depth: Integer read FDepth;
    property Size: UInt64 read FSize write FSize;
  end;

  TSquarifiedTreeMap = class
  private
    FMargeText: integer;
    FMArgeBox: integer;
    FRoot: TFileNode;
    FBitmap: TBitmap;
    procedure DrawNode(ANode: TFileNode);
    procedure Squarify(ANode: TFileNode; const ARect: TRect);
    procedure DrawChildren(ANode: TFileNode);
    function Reduce(AR: TRect): TRect;
  public
    constructor Create(const AW, AH: Integer);
    destructor Destroy; override;

    procedure BuildTreeMap(const ADirectory: string);
    procedure DrawTreeMap;
    function GetBitmap: TBitmap;
  end;

implementation

uses
  System.StrUtils,
  System.UITypes;

procedure RecursivelyAddFiles(ANode: TFileNode; const ADirectory: string; ADepth: Integer);
var
  Files: TStringDynArray;
  FileSize: UInt64;
  ChildNode: TFileNode;
begin
  Files := TDirectory.GetFiles(ADirectory);

  for var FilePath in Files do
  begin
    FileSize  := TFile.GetSize(FilePath);
    ChildNode := TFileNode.Create(FilePath, FileSize, ADepth);

    ANode.Children.Add(ChildNode);

    // Ajouter la taille du fichier au parent
    ANode.Size := ANode.Size + FileSize;
  end;

  for var SubDir in TDirectory.GetDirectories(ADirectory) do
  begin
    ChildNode := TFileNode.Create(SubDir, 0, ADepth);

    ANode.Children.Add(ChildNode);

    RecursivelyAddFiles(ChildNode, SubDir, ADepth + 1);

    // Ajouter la taille du fichier au parent
    ANode.Size := ANode.Size + ChildNode.Size;
  end;
end;

{ TFileNode }
constructor TFileNode.Create(const AFileName: string; ASize: Int64;
  ADepth: Integer);
begin
  FFileName := AFileName;
  FSize     := ASize;
  FDepth    := ADepth;
  Children  := TObjectList<TFileNode>.Create;
end;

destructor TFileNode.Destroy;
begin
  Children.Free;

  inherited;
end;

procedure TFileNode.ShowArbo;

  function ShowChildren(ANode: TFileNode): string;
  const
    CHAINE = 'N: %32s S: %.4d C: %.1d L: %.3d T: %.3d W: %.3d H: %.3d';
  var
    N: TFileNode;
  begin
    Result := Format(CHAINE, [ANode.FileName, ANode.Size, ANode.Children.Count, ANode.Bloc.Left, ANode.Bloc.Top, ANode.Bloc.Width, ANode.Bloc.Height]) + #13#10;

    if ANode.Children.Count > 0 then
      for N in ANode.Children do
        Result := Result + ShowChildren(N);
  end;

begin
  ShowMessage(ShowChildren(Self));
end;

{ TSquarifiedTreeMap }
constructor TSquarifiedTreeMap.Create(const AW, AH: Integer);
begin
  Randomize;

  FRoot                     := nil;
  FBitmap                   := TBitmap.Create(AW, AH);
  FBitmap.Canvas.Font.Color := clWhite;
  FBitmap.Canvas.Font.Size  := 6;
  FMargeBox                 := 3;
  FMargeText                := 1 + FBitmap.Canvas.TextHeight('Deren');
end;

destructor TSquarifiedTreeMap.Destroy;
begin
  FRoot.Free;
  FBitmap.Free;

  inherited;
end;

procedure TSquarifiedTreeMap.Squarify(ANode: TFileNode; const ARect: TRect);
var
  Child: TFileNode;
  CHildRect: TRect;
  TotalSize: Int64;
  Sum, Ratio: Extended;
  Width, Height: Integer;
begin
  ANode.Bloc := ARect;
  CHildRect := Reduce(ARect);

  if ANode.Children.Count > 0 then
  begin
    Width  := CHildRect.Width;
    Height := CHildRect.Height;

    // Vrification pour viter Les erreurs d'affichages.
    if (Width > 0) and (Height > 0) then
    begin
      TotalSize := ANode.Size;
      Sum       := 0;

      if Height > Width then
        Ratio := TotalSize / Height
      else
        Ratio := TotalSize / Width;

      for Child in ANode.Children do
      begin
        Child.Bloc := CHildRect;

        if Height > Width then
        begin
          Child.Bloc.Top    := CHildRect.Top + Round(Sum);
          Child.Bloc.Height := Round(Child.Size / Ratio);
          Sum               := Sum + Child.Bloc.Height;
        end
        else
        begin
          Child.Bloc.Left  := CHildRect.Left + Round(Sum);
          Child.Bloc.Width := Round(Child.Size / Ratio);
          Sum              := Sum + Child.Bloc.Width;
        end;

        Squarify(Child, Child.Bloc);
      end;
    end;
  end;
end;

procedure TSquarifiedTreeMap.BuildTreeMap(const ADirectory: string);
begin
  FRoot := TFileNode.Create(ADirectory, 0, 0);

  RecursivelyAddFiles(FRoot, ADirectory, 1);

  Squarify(FRoot, FBitmap.Canvas.ClipRect);

  DrawTreeMap;

//  FRoot.ShowArbo;
end;

procedure TSquarifiedTreeMap.DrawNode(ANode: TFileNode);
var
  i, T: Integer;
  C: TColor;
begin
  if ANode.Size = 0 then
    Exit;

  C := TColor(Random($FFFFFF));

  FBitmap.Canvas.Brush.Color := C;
  FBitmap.Canvas.Pen.Color   := C;

  FBitmap.Canvas.Rectangle(ANode.Bloc);

  if ANode.Bloc.Height > FMargeText then
  begin
    T := FBitmap.Canvas.TextWidth(ANode.FileName);

    if ANode.Bloc.Width > T then
    begin
      i := (ANode.Bloc.Width - T) div 2;

      FBitmap.Canvas.TextOut(ANode.Bloc.Left + i, ANode.Bloc.Top, ANode.FileName);
    end;
  end;
end;

procedure TSquarifiedTreeMap.DrawTreeMap;
begin
  if Assigned(FRoot) then
  begin
    DrawNode(FRoot);

    DrawChildren(FRoot);
  end;
end;

function TSquarifiedTreeMap.GetBitmap: TBitmap;
begin
  Result := FBitmap;
end;

procedure TSquarifiedTreeMap.DrawChildren(ANode: TFileNode);
var
  Child: TFileNode;
begin
  for Child in ANode.Children do
  begin
    DrawNode(Child);

    DrawChildren(Child);
  end;
end;

function TSquarifiedTreeMap.Reduce(AR: TRect): TRect;
begin
  Result.Left   := AR.Left   + FMargeBox;
  Result.Top    := AR.Top    + FMargeText;
  Result.Right  := AR.Right  - FMargeBox;
  Result.Bottom := AR.Bottom - FMargeBox;
end;

end.
